package org.netkernel.layer1.endpoint;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.xml.parsers.ParserConfigurationException;

import org.netkernel.container.IKernel;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.request.IExpiry;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.IRequestor;
import org.netkernel.request.IResponse;
import org.netkernel.request.IResponseMeta;
import org.netkernel.request.impl.ExpiryFactory;
import org.netkernel.request.impl.MetaImpl;
import org.netkernel.request.impl.RequestCloneImpl;
import org.netkernel.request.impl.RequestResponseFieldsImpl;
import org.netkernel.request.impl.ResponseImpl;
import org.netkernel.scheduler.IInternalResponseMeta;
import org.netkernel.urii.ISpace;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class JavaRuntimeAccessor extends StandardAccessorImpl
{
	private Element mGrammar;
	private Map<CachedEndpoint, CachedEndpoint> mEndpointCache = new ConcurrentHashMap<CachedEndpoint, CachedEndpoint>();	
	
	public JavaRuntimeAccessor() throws ParserConfigurationException
	{
		declareThreadSafe();
		
		Document d=XMLUtils.newDocument();
		Element g=d.createElement("grammar");
		d.appendChild(g);
		Element a=d.createElement("active");
		g.appendChild(a);
		Element t=d.createElement("identifier");
		XMLUtils.setText(t, "active:java");
		a.appendChild(t);
		Element v=d.createElement("varargs");
		a.appendChild(v);
		mGrammar=g;
	}
	
	public void onCommissionEndpoint(IKernel aKernel, ISpace aSpace) throws Exception
	{	super.onCommissionEndpoint(aKernel, aSpace);
	}
	
	public void onDecommissionEndpoint() throws Exception
	{	super.onDecommissionEndpoint();
	}	

	public void onRequest(INKFRequestContext aContext) throws Exception
	{
		int scopeDepth=Integer.MAX_VALUE;
		if (aContext.getThisRequest().getVerb()==INKFRequestReadOnly.VERB_META)
		{	super.onMeta(aContext);
		}
		else
		{	String className = aContext.getThisRequest().getArgumentValue("class");
			IStandardEndpoint ep;
			Class c=null;
			ISpace space=null;
			IRequestScopeLevel scope=aContext.getKernelContext().getThisKernelRequest().getRequestScope().getParent();
			while(c==null && scope!=null)
			{	space=scope.getSpace();
				ClassLoader cl=space.getClassLoader();
				if (cl!=null)
				{	try
					{	c=cl.loadClass(className);
						scopeDepth=scope.getDepth();
					}
					catch (ClassNotFoundException e)
					{
					}
				}
				scope=scope.getParent();
			}
			if (c==null)
			{	throw new ClassNotFoundException(className);
			}
			else
			{	CachedEndpoint cepKey=new CachedEndpoint(className, space, null);
				CachedEndpoint cep=mEndpointCache.get(cepKey);
				if (cep!=null)
				{	ep=cep.getEndpoint();
				}
				else
				{	try
					{	ep = (IStandardEndpoint)c.newInstance();
						ep.setParameterRaw("grammar", mGrammar, getKernel());
						ep.onCommissionEndpoint(aContext.getKernelContext().getKernel(), this.getSpace());
						cep=new CachedEndpoint(className, space, ep);
						mEndpointCache.put(cep, cep);
					}
					catch (ClassCastException e)
					{	throw aContext.createFormattedException("EX_JAVA_NOT_ACCESSOR", "MSG_JAVA_NOT_ACCESSOR", null, null, c.getName());
					}
				}
			}
				
			aContext.setNoResponse(); //endpoint will do the response
			IRequest request=aContext.getKernelContext().getThisKernelRequest();
			if (scopeDepth<Integer.MAX_VALUE)
			{	final RequestorProxy requestorProxy=new RequestorProxy(scopeDepth,request);
				final IRequest parentRequest=request;
				request=new RequestCloneImpl(request)
				{	public IRequest getParent()
					{	return parentRequest;
					}
					public IRequestor getRequestor()
					{	return requestorProxy;
					}
				};
			}
			ep.onAsyncRequest(request);
		}
	}
	
	private static class RequestorProxy implements IRequestor
	{	private final int mRequestDepth;
		private final IRequest mRequest;
		public RequestorProxy(int aRequestDepth, IRequest aRequest)
		{	mRequestDepth=aRequestDepth;
			mRequest=aRequest;
		}
		public void onResponse(IResponse aResponse)
		{	
			//TODO this code creates a lot of coupling to internals of kernel/nkf (should really be factored out)
			IExpiry[] expDet=new IExpiry[]{ExpiryFactory.getNeverExpired()};
			IInternalResponseMeta m = IInternalResponseMeta.create(ExpiryFactory.getNeverExpired(), RequestResponseFieldsImpl.EMPTY, mRequestDepth, mRequestDepth, 0,0, new IInternalResponseMeta[0],0,expDet);
			IResponse expiryDep = new ResponseImpl(null,null,m);
			
			IRequestResponseFields fieldsIn=aResponse.getMeta().getUserMetaData();
			List<IResponse> additionalDependencies=(List<IResponse>)fieldsIn.getValue(IResponseMeta.KEY_ADDITIONAL_DEPENDENCIES);
			if (additionalDependencies!=null)
			{	additionalDependencies.add(expiryDep);
			}
			else
			{	additionalDependencies=new ArrayList<IResponse>(1);
				additionalDependencies.add(expiryDep);
				RequestResponseFieldsImpl fields=new RequestResponseFieldsImpl(fieldsIn);
				fields.add(IResponseMeta.KEY_ADDITIONAL_DEPENDENCIES, additionalDependencies);
				fieldsIn=fields;
			}
			IResponseMeta meta=new MetaImpl(aResponse.getMeta().getExpiry(),fieldsIn);
			IResponse responseOut=new ResponseImpl(mRequest, aResponse.getRepresentation(), meta);
			mRequest.getRequestor().onResponse(responseOut);
		}
	}
	
	/** Holder class around endpoint instance in case we want to hold LRU etc in future
	 * @author tab
	 *
	 */
	private static class CachedEndpoint
	{	private IStandardEndpoint mEndpoint;
		private String mName;
		private ISpace mSpace;
	
		public CachedEndpoint(String aName, ISpace aSpace, IStandardEndpoint aEndpoint)
		{	mEndpoint=aEndpoint;
			mName=aName;
			mSpace=aSpace;
		}
		
		public boolean equals(Object aOther)
		{	if (aOther instanceof CachedEndpoint)
			{	CachedEndpoint other=(CachedEndpoint)aOther;
				return mName.equals(other.mName) && mSpace==other.mSpace;
			}
			else
			{	return false;
			}
		}
		
		public int hashCode()
		{	return mSpace.hashCode()^mName.hashCode();
		}
		
		public IStandardEndpoint getEndpoint()
		{	return mEndpoint;
		}
	}
	
}