package org.netkernel.layer1.endpoint;

import java.util.*;

import org.netkernel.layer0.nkf.INKFExpiryFunction;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSPredicateFactory;
import org.netkernel.layer0.util.GoldenThreadExpiryFunction;
import org.netkernel.layer0.util.PairList;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;
import org.netkernel.urii.impl.Version;

public class SimpleImportDiscoveryAccessor extends StandardAccessorImpl
{
    private GoldenThreadExpiryFunction mExpiry;
    private long mInitDelay;
    private String mType;
    
    public SimpleImportDiscoveryAccessor()
    {    this.declareThreadSafe();
    }
    
    public void postCommission(INKFRequestContext aContext) throws Exception
    {  try
    	{	mInitDelay=aContext.getKernelContext().getKernel().getConfiguration().getLong("netkernel.simpleimportdiscovery.init.delay");
    	}
    	catch(Exception e)
    	{ ; } //No property set
        mType=(String)aContext.getParamValue("type");
        if(mInitDelay > 0)
        {    mExpiry=new GoldenThreadExpiryFunction("SID delay");
            Thread t=new Thread()
            {
                public void run()
                {    try
                    {    Thread.sleep(mInitDelay);
                    } catch (Exception e) {;}
                    System.out.println("Activating Delayed SimpleDynamicImport "+mType);
                    mExpiry.invalidate();
                    mExpiry = null;
                }
            };
            t.start();
        }
    }
    
    public void onSource(INKFRequestContext aContext) throws NKFException
    {
        INKFExpiryFunction expiry=mExpiry;
        if (expiry==null)
        {
            String type=(String)this.getParameter("type");
            String uri=(String)this.getParameter("uri");
           
            Set<ISpace> superStack=new HashSet();
            IRequestScopeLevel scope=aContext.getKernelContext().getThisKernelRequest().getRequestScope();
            while (scope!=null)
            {    ISpace s=scope.getSpace();
                superStack.add(s);
                scope=scope.getParent();
            }
           
            aContext.logFormatted(INKFRequestContext.LEVEL_FINE, "LOG_DYNAMIC_IMPORT_TYPE", uri, type);
           
            INKFRequest req=aContext.createRequest("active:spaceAggregateHDS");
            req.addArgument("uri", uri);
            req.addArgumentByValue("exclusions", superStack);
            req.setRepresentationClass(IHDSNode.class);
            IHDSNode n=(IHDSNode)aContext.issueRequest(req);
            IHDSNodeList nl=n.getNodes("/spaces/space").filter(HDSPredicateFactory.hasChild("connection"));
            List<DISortStruct> imports = new ArrayList<SimpleImportDiscoveryAccessor.DISortStruct>();
            boolean needsSort=false;
            for(IHDSNode m : nl)
            {	for(IHDSNode m2: m.getNodes("connection/type"))
                {	String t = m2.getValue().toString();
                    if(type.equals(t))
                    {	String id=(String)m.getFirstValue("id");
                        String version=(String)m.getFirstValue("version");
                        String priority=(String)m.getFirstValue("connection/priority");
                        boolean add=true;
                        for (int i=0; i<imports.size(); i++)
                        {	DISortStruct diss2=imports.get(i);
                        	if (diss2.mId.equals(id))
                        	{	Version v1=new Version(diss2.mVersion);
                        		Version v2=new Version(version);
                        		if(v2.isGreaterThan(v1))
                        		{	//replace existing older version
                        			imports.remove(i);
                        			aContext.logFormatted(INKFRequestContext.LEVEL_FINE, "LOG_DYNAMIC_IMPORT_IGNORE_WARN", diss2.mId, diss2.mVersion);
                        		}
                        		else
                        		{	//ignore this version
                        			add=false;
                        			aContext.logFormatted(INKFRequestContext.LEVEL_FINE, "LOG_DYNAMIC_IMPORT_IGNORE_WARN", id, version);
                        		}
                        		break;
                        	}
                        }

                        if (add)
                        {	DISortStruct diss=new DISortStruct(id, version, priority);
                        	imports.add(diss);
                        	if (diss.mPriority!=DISortStruct.DEFAULT_PRIORITY)
                        	{	needsSort=true;
                        	}
                        }
                        
                        break;
                    }
                }
            }
            
            if (needsSort)
            {	Collections.sort(imports);
            }
            
            HDSBuilder b = new HDSBuilder();
            b.pushNode("config");
            for (DISortStruct diss : imports)
            {	aContext.logFormatted(INKFRequestContext.LEVEL_FINE, "LOG_DYNAMIC_IMPORT", diss.mId, diss.mVersion);
            	b.pushNode("space");
                b.addNode("uri",diss.mId);
                b.addNode("version",diss.mVersion);
                b.popNode();
            }
            
            if(imports.size()==0)
            {    aContext.logFormatted(INKFRequestContext.LEVEL_FINER, "LOG_DYNAMIC_IMPORT_TYPE_NOT_FOUND", uri, type);
            }
            aContext.createResponseFrom(b.getRoot());
        }
        else
        {	//dummy response
            HDSBuilder b = new HDSBuilder();
            b.pushNode("config");
            aContext.createResponseFrom(b.getRoot()).setExpiry(INKFResponse.EXPIRY_FUNCTION,expiry);
        }
    }
    
    private static class DISortStruct implements Comparable<DISortStruct>
    {	public String mId;
    	public String mVersion;
    	public float mPriority;
    	
    	public static final float DEFAULT_PRIORITY=1.0f;
    	
    	public DISortStruct(String aId, String aVersion, String aPriority)
    	{	mId=aId;
    		mVersion=aVersion;
    		mPriority=DEFAULT_PRIORITY;
    		try
    		{	if (aPriority!=null)
    			{	mPriority=Float.parseFloat(aPriority);
    			}
    		} catch (Exception e) {;}
    	}

		public int compareTo(DISortStruct o)
		{	float diff=o.mPriority-mPriority;
			if (diff==0)
			{	return 0;
			}
			else
			{	return (diff>0)?1:-1;
			}
		}
    }
}

