package org.netkernel.layer1.endpoint;

import org.netkernel.container.IKernel;
import org.netkernel.container.ISpaceListener;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.nkf.impl.NKFResponseReadOnlyImpl;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.urii.AggregatingPolledExpiry;
import org.netkernel.layer0.util.GoldenThreadExpiryFunction;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.request.IExpiry;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;

import java.util.*;
import java.util.concurrent.locks.ReentrantLock;

public class SpaceHDSAggregator extends StandardAccessorImpl implements ISpaceListener
{
	private GoldenThreadExpiryFunction mSpaceChangeExpiry;
	
	private ReentrantLock mCachabilityTestLock=new ReentrantLock(true);
	private Map<String,Long> mLastRequested=new HashMap<String, Long>();
	private String mMonitoredURI;
	private Map<String,IExpiry> mMonitoredExpiries;	
	
	public SpaceHDSAggregator()
	{	this.declareThreadSafe();
		this.declareSourceRepresentation(IHDSNode.class);
		spaceChanged();
	}
	
	public void onCommissionEndpoint(IKernel aKernel, ISpace aSpace) throws Exception
	{	super.onCommissionEndpoint(aKernel, aSpace);
		aKernel.addSpaceListener(this);
	}
	
	public void onDecommissionEndpoint() throws Exception
	{	getKernel().removeSpaceListener(this);
		super.onDecommissionEndpoint();
	}
	
	public void spaceChanged()
	{	if (mSpaceChangeExpiry!=null)
		{	mSpaceChangeExpiry.invalidate();
		}
		mSpaceChangeExpiry=new GoldenThreadExpiryFunction("SpaceHDSAggregator.ISpaceListener");
	}
	
	@Override
	public void onSource(INKFRequestContext context) throws Exception
	{	
		Set<ISpace> superStack=null;
		if (context.getThisRequest().argumentExists("exclusions"))
		{	superStack=(Set)context.source("arg:exclusions",Set.class);
		}
		
		IKernel kernel=context.getKernelContext().getKernel();
		HDSBuilder b=new HDSBuilder();
		b.pushNode("spaces");
		String uri=context.getThisRequest().getArgumentValue("uri");
		b.addNode("uri", uri);
		Iterator<ISpaceWithIdentity> it=kernel.getSpaces();
		boolean errors=false;
		boolean expired=false;
		
		long now=System.currentTimeMillis();
		boolean tc=performNonCachabilityTest(uri,now,context);
		
		List<IExpiry> expiries=new ArrayList<IExpiry>();
		Map<String,IExpiry> expiriesBySpace=null;
		if (tc)
		{	expiriesBySpace=new HashMap<String, IExpiry>();
		}
		while(it.hasNext())
		{	ISpaceWithIdentity space=it.next();
			b.pushNode("space");
			String id=space.getIdentifier().toString();
			b.addNode("id", id);
			b.addNode("version", space.getVersion().toString());
			
			if (superStack!=null)
			{	if (superStack.contains(space))
				{	b.popNode();
					continue;
				}
			}
			
			IRequestScopeLevel scope=RequestScopeLevelImpl.createOrphanedRootScopeLevel(space, context.getKernelContext().getRequestScope());
			INKFRequest req=context.createRequest(uri);
			req.setRequestScope(scope);
			Object rep;
			try
			{	INKFResponseReadOnly resp=context.issueRequestForResponse(req);
				IExpiry expiry=((NKFResponseReadOnlyImpl)resp).getMeta().getExpiry();
				boolean exp=expiry.isExpired(now);
				rep=resp.getRepresentation();
				if (tc && rep!=null)
				{	expiriesBySpace.put(id, expiry);
				}
				expired|=exp;
				expiries.add(expiry);
			}
			catch(Exception e)
			{	//Space doesn't contain the resource
				rep=null;
			}
			if (rep!=null)
			{	
				try
				{	if (!(rep instanceof IHDSNode))
					{	rep=(IHDSNode)context.transrept(rep, IHDSNode.class);
					}
					b.importChildren(((IHDSNode)rep));
				}
				catch (NKFException e)
				{	context.logFormatted(INKFLocale.LEVEL_WARNING, "MSG_SA_IMP_BAD_TYPE", rep.getClass().getName(),space.getIdentifier().toString(),uri);
					errors=true;
				}
			}
			b.popNode();
		}
		b.popNode();
		
		if (tc)
		{	mMonitoredExpiries=expiriesBySpace;
		}
		
		
		INKFResponse resp=context.createResponseFrom(b.getRoot());
		if (errors)
		{	resp.setExpiry(INKFResponse.EXPIRY_FUNCTION,mSpaceChangeExpiry);
		}
		else
		{	if (expired)
			{	long expiryTime=getKernel().getTimer().getApproximateTime()+getKernel().getConfiguration().getLong("stdmod.initialisation.hysteresis", 500L);
				resp.setExpiry(INKFResponse.EXPIRY_CONSTANT, expiryTime);
			}
			else
			{	expiries.add(mSpaceChangeExpiry);
				long poll=getKernel().getConfiguration().getLong("netkernel.poll.internal", 500L);
				String debug="SpaceHDSAggregator("+uri+")";
				resp.setExpiry(INKFResponse.EXPIRY_FUNCTION,new AggregatingPolledExpiry(expiries,poll,debug));
			}
		}
	}
	
	private boolean performNonCachabilityTest(String aURI, long aNow, INKFRequestContext aContext)
	{	mCachabilityTestLock.lock();
		boolean result=false;
		try
		{	Long lastExecuted=mLastRequested.put(aURI, aNow);
			if (lastExecuted!=null)
			{	long diff=aNow-lastExecuted;
				if (diff<2000) //if it gets re-executed more often than every x ms
				{	if (mMonitoredURI==null)
					{	mMonitoredURI=aURI;
						result=true;
					}
					else if (mMonitoredURI.equals(aURI) )
					{	
						Map<String,IExpiry> monitoredExpiries=mMonitoredExpiries;
						if (monitoredExpiries!=null)
						{
							for (Map.Entry<String, IExpiry> entry : monitoredExpiries.entrySet())
							{	if (entry.getValue().isExpired(aNow))
								{	aContext.logFormatted(INKFLocale.LEVEL_WARNING, "MSG_SA_EXPIRE", aURI,entry.getKey());
								}
							}
						}
						mMonitoredExpiries=null;
						mMonitoredURI=null;
					}
				}
			}
		}
		catch (Exception e)
		{	e.printStackTrace();
		}
		finally
		{	mCachabilityTestLock.unlock();
		}
		return result;
	}
}
