package org.netkernel.layer1.endpoint;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;

import org.netkernel.layer0.meta.IEndpointStateMeta;
import org.netkernel.layer0.nkf.INKFAsyncRequestHandle;
import org.netkernel.layer0.nkf.INKFAsyncRequestListener;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.module.standard.endpoint.TransparentOverlayImpl;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestResponseFieldsImpl;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.w3c.dom.Document;


public class ThrottleOverlay extends TransparentOverlayImpl implements INKFAsyncRequestListener, IEndpointStateMeta
{		
	final private ThrottleState mThrottle=new ThrottleState();
	private static final int DEFAULT_CONCURRENCY=8;
	private static final int DEFAULT_QUEUE=250;
	private int mConcurrency=DEFAULT_CONCURRENCY;
	private int mQueue=DEFAULT_QUEUE;
	private INKFResponseReadOnly<Document> mConfigurationResponse;
	private AtomicBoolean mReconfiguring=new AtomicBoolean(false);

	public ThrottleOverlay() throws Exception
	{	declareThreadSafe();
		declareUnhandledExceptionsExpired(false);
	}

	protected void postCommission(INKFRequestContext context) throws Exception
	{	checkIfReconfigureRequired(context);
	}
	
	private void checkIfReconfigureRequired(INKFRequestContext context) throws Exception
	{	if( (mConfigurationResponse==null || mConfigurationResponse.isExpired()) && mReconfiguring.compareAndSet(false, true))
		{	int oldQueue=mQueue;
			int oldConcurrency=mConcurrency;
			try
			{	doConfigure(context);
			}
			catch(NKFException e)
			{	context.logRaw(INKFRequestContext.LEVEL_WARNING, "Error sourcing configuration. Settings unchanged:"+e.getDeepestId()+" "+e.getDeepestMessage());
				mQueue=oldQueue;
				mConcurrency=oldConcurrency;
			}
			finally
			{	mReconfiguring.set(false);				
			}
		}
	}
	
	private void doConfigure(INKFRequestContext context) throws Exception
	{	
		INKFResponseReadOnly<Document> resp=context.sourceForResponse("param:config", Document.class);
		Document config=resp.getRepresentation();
		if(config!=null)
		{	XMLReadable c=new XMLReadable(config);
			String s=c.getText("/config/concurrency");
			if(!"".equals(s))
			{	mConcurrency=Integer.parseInt(s);
				if(mConcurrency<=0)
				{	context.logRaw(INKFRequestContext.LEVEL_WARNING, "Throttle concurrency must be greater than zero - using default of 8");
					mConcurrency=DEFAULT_CONCURRENCY;
				}
			}
			s=c.getText("/config/queue");
			if(!"".equals(s))
			{	mQueue=Integer.parseInt(s);
				if(mQueue<0)
				{	context.logRaw(INKFRequestContext.LEVEL_WARNING, "Throttle queue must be greater than or equal to zero - using default of 250");
					mQueue=DEFAULT_QUEUE;
				}
			}
		}
		context.logRaw(INKFRequestContext.LEVEL_INFO, "Throttle configured: concurrency="+mConcurrency+" queue="+mQueue);
		mConfigurationResponse=resp;
	}

	public void onRequest(String aID, INKFRequestContext aContext) throws Exception
	{	
		if (aContext.getThisRequest().getVerb()==INKFRequestReadOnly.VERB_META)
		{	Utils.delegateRequestInto(getDelegateSpace(),aContext);
		}
		else
		{
			// decide what to do execute now, queue or fail
			boolean issueNow=false;
			synchronized(mThrottle)
			{	if (mThrottle.mIssuedRequests<mConcurrency)
				{	// execute now
					mThrottle.mIssuedRequests++;
					issueNow = true;
				}
				else if (mThrottle.mList.size()<mQueue)
				{	// queue request
					mThrottle.mList.add(aContext);
				}
				else
				{	// fail fast
					throw new NKFException("Throttle Overload","Queued requests exceeded maximum, request rejected", null);
				}
			}
			// issue request if possible
			if (issueNow)
			{	issueSubRequest(aContext);
			}
			checkIfReconfigureRequired(aContext);
			aContext.setNoResponse();  //Force no response yet.
		}
	}

	private void issueSubRequest(INKFRequestContext aContext) throws NKFException
	{	INKFRequestReadOnly reqIn=aContext.getThisRequest();
		INKFRequest reqOut=aContext.createRequest(reqIn.getIdentifier());
		reqOut.setVerb(reqIn.getVerb());
		reqOut.setRepresentationClass(reqIn.getRepresentationClass());
		INKFResponseReadOnly primaryResp=reqIn.getPrimaryAsResponse();
		if (primaryResp!=null)
		{	reqOut.addPrimaryArgumentFromResponse(primaryResp);
		}
		IRequestScopeLevel rctx = RequestScopeLevelImpl.createNonDurableScopeLevel(this.getDelegateSpace(), aContext.getKernelContext().getThisKernelRequest().getRequestScope());
		reqOut.setRequestScope(rctx);
		INKFAsyncRequestHandle handle= aContext.issueAsyncRequest(reqOut);
		handle.setListener(this);
		
	}
	
	/** called when sub-request completes with exception
	 */
	public void receiveException(NKFException exception, INKFRequest aRequest, INKFRequestContext aContext) throws Exception
	{	// release potential queued requests
		assessThrottle();
		// propage exception up
		throw exception;
	}
	
	public void receiveResponse(INKFResponseReadOnly response, INKFRequestContext aContext) throws Exception
	{	// release potential queued requests
		assessThrottle();	
		// propagate result
		aContext.createResponseFrom(response);
	}
	
	/** pull off a queued request if one is available and issue it
	 */
	private void assessThrottle()
	{	boolean retry=true;
		while(retry)
		{	INKFRequestContext aContext;
			synchronized(mThrottle)
			{	if (mThrottle.mList.size()>0)
				{	aContext=(INKFRequestContext)mThrottle.mList.remove(0);
				}
				else
				{	aContext=null;
					mThrottle.mIssuedRequests--;
				}
			}
			try
			{	if (aContext!=null)
				{	issueSubRequest(aContext);
				}
				retry=false;
			}
			catch (NKFException e)
			{	//TODO!
                // we cannot return errors in the execution of this request using NKF
				// we must report them back to the client of this request
				// This situation should never happen if the context is well-formed.
				// However NKF doesn't support this so we must go low-level

				// Need Mr B's solution here...
				
				// try next item
				retry=true;
			}
		}
	}
	
	@Override
	public IRequestResponseFields getState()
	{
		RequestResponseFieldsImpl fields=new RequestResponseFieldsImpl(super.getState());
		fields.put("concurrency", mConcurrency);
		fields.put("admitted", mThrottle.mIssuedRequests);
		fields.put("max-queue", mQueue);
		fields.put("queued", mThrottle.mList.size());
		return fields;
	}



	
	/** Throttle state class
	 **/
	private static class ThrottleState
	{	/** number of requests currently in execution state*/
		public int mIssuedRequests = 0;
		/** list of requests pending execution */
		public final List mList = new ArrayList();
	}
}